from typing import Tuple, Union, Literal, TypeVar, Iterable, Optional, Sequence

from qm.qua._expressions import Scalar, QuaArrayVariable
from qm.grpc.qua import (
    QuaProgramRampPulse,
    QuaProgramVarRefExpression,
    QuaProgramAnyScalarExpression,
    QuaProgramArrayVarRefExpression,
)

ChirpUnits = Literal[
    "Hz/nsec",
    "GHz/sec",
    "mHz/nsec",
    "MHz/sec",
    "uHz/nsec",
    "KHz/sec",
    "nHz/nsec",
    "Hz/sec",
    "pHz/nsec",
    "mHz/sec",
]
ChirpType = Union[
    Tuple[Union[Iterable[int], QuaArrayVariable[int], Scalar[int]], ChirpUnits],
    Tuple[Iterable[int], Iterable[int], ChirpUnits],
]
"""A type for performing piecewise linear sweep of the element’s intermediate frequency in time.
A tuple, with the 1st element being a list of rates and the second should be a string with the unit type.
See the ChirpUnits type for the complete list of supported units.
"""

MessageExpressionType = QuaProgramAnyScalarExpression
AmpValuesType = Tuple[
    MessageExpressionType,
    Optional[MessageExpressionType],
    Optional[MessageExpressionType],
    Optional[MessageExpressionType],
]
MeasurePulseType = Union[str, Tuple[str, AmpValuesType]]
PlayPulseType = Union[MeasurePulseType, QuaProgramRampPulse]
MessageArrayVarType = QuaProgramArrayVarRefExpression
MessageVarType = QuaProgramVarRefExpression

T = TypeVar("T")
OneOrMore = Union[T, Sequence[T]]

ConvolutionMode = Literal["", "valid", "same", "full"]



import abc
from typing import TYPE_CHECKING, List, Tuple

from betterproto.lib.google.protobuf import Value, ListValue

from qm.grpc.qua import QuaResultAnalysis

if TYPE_CHECKING:
    from qm.qua._dsl import _ResultStream


class _OutputStream(metaclass=abc.ABCMeta):
    """Even though it looks like a stream, it does not support operations (like __add__) and hence, a different object"""

    def __init__(self, input_stream: "_ResultStream", tag: str):
        self._input_stream = input_stream
        self.tag = tag

    def to_proto(self) -> "ListValue":
        values = [Value(string_value=s) for s in self._operator_array]
        values.append(self._input_stream.to_proto())
        return ListValue(values=values)

    @property
    @abc.abstractmethod
    def _operator_array(self) -> Tuple[str, ...]:
        pass


class _SaveOutputStream(_OutputStream):
    @property
    def _operator_array(self) -> Tuple[str, ...]:
        return "save", self.tag


class _SaveAllOutputStream(_OutputStream):
    @property
    def _operator_array(self) -> Tuple[str, ...]:
        return "saveAll", self.tag


class _AutoSaveAllOutputStream(_OutputStream):
    @property
    def _operator_array(self) -> Tuple[str, ...]:
        return "saveAll", self.tag, "auto"


class _ResultAnalysis:
    def __init__(self, result_analysis: QuaResultAnalysis):
        self._result_analysis = result_analysis
        self._saves: List[_OutputStream] = []

    def _add_output_stream(self, output_stream: _OutputStream) -> None:
        for save in self._saves:
            if save.tag == output_stream.tag:
                raise Exception("can not save two streams with the same tag")
        self._saves.append(output_stream)

    def save(self, tag: str, expression: "_ResultStream") -> None:
        self._add_output_stream(_SaveOutputStream(expression, tag))

    def save_all(self, tag: str, expression: "_ResultStream") -> None:
        self._add_output_stream(_SaveAllOutputStream(expression, tag))

    def auto_save_all(self, tag: str, expression: "_ResultStream") -> None:
        self._add_output_stream(_AutoSaveAllOutputStream(expression, tag))

    def _add_pipeline(self, output: _OutputStream) -> None:
        proto_output = output.to_proto()
        self._result_analysis.model.append(proto_output)

    def generate_proto(self) -> None:
        for save in self._saves:
            self._add_pipeline(save)



from pathlib import Path
from typing import Set, Union, Optional, FrozenSet

from qm import DictQuaConfig
from qm.grpc.qua_config import QuaConfig
from qm.api.models.capabilities import Capability
from qm.program._qua_config_schema import load_config
from qm.program._ResultAnalysis import _ResultAnalysis
from qm.program.StatementsCollection import StatementsCollection
from qm.grpc.qua import (
    QuaProgram,
    QuaProgramScript,
    QuaResultAnalysis,
    QuaProgramVarDeclaration,
    QuaProgramStatementsCollection,
)


class Program:
    def __init__(
        self,
        config: Optional[QuaConfig] = None,
        program: Optional[QuaProgram] = None,
    ):
        if program is None:
            program = QuaProgram(
                script=QuaProgramScript(variables=[], body=QuaProgramStatementsCollection(statements=[])),
                result_analysis=QuaResultAnalysis(model=[]),
            )

        self._program = program
        self._qua_config = config
        self._result_analysis = _ResultAnalysis(self._program.result_analysis)
        self._is_in_scope = False
        self._used_capabilities: Set[Capability] = set()

    def add_declaration(self, declaration: QuaProgramVarDeclaration) -> None:
        self._program.script.variables.append(declaration)

    @property
    def body(self) -> StatementsCollection:
        return StatementsCollection(self._program.script.body)

    @property
    def result_analysis(self) -> _ResultAnalysis:
        return self._result_analysis

    @property
    def qua_program(self) -> QuaProgram:
        return self._program

    def to_protobuf(self, config: DictQuaConfig) -> bytes:
        """
        Serialize the program to a protobuf binary.
        """
        loaded_config = load_config(config)
        copy = QuaProgram().from_dict(self._program.to_dict())
        copy.config = QuaConfig().from_dict(loaded_config.to_dict())
        return bytes(copy)

    @classmethod
    def from_protobuf(cls, binary: bytes) -> "Program":
        """
        Deserialize the program from a protobuf binary.
        """
        program = QuaProgram().parse(binary)
        return cls(program=program, config=program.config)

    def to_file(self, path: Union[str, Path], config: DictQuaConfig) -> None:
        """
        Serialize the program to a protobuf binary and write it to a file.
        """
        if isinstance(path, str):
            path = Path(path)
        path.write_bytes(self.to_protobuf(config))

    @classmethod
    def from_file(cls, path: Union[str, Path]) -> "Program":
        """
        Deserialize the program from a protobuf binary file.
        """
        if isinstance(path, str):
            path = Path(path)
        return cls.from_protobuf(path.read_bytes())

    def set_in_scope(self) -> None:
        self._is_in_scope = True

    def set_exit_scope(self) -> None:
        self._is_in_scope = False

    def is_in_scope(self) -> bool:
        return self._is_in_scope

    @property
    def used_capabilities(self) -> FrozenSet[Capability]:
        return frozenset(self._used_capabilities)

    def add_used_capability(self, capability: Capability) -> None:
        self._used_capabilities.add(capability)



import json
import math
import time
import asyncio
import logging
import functools
from abc import ABCMeta, abstractmethod
from typing_extensions import ParamSpec
from typing import Type, Generic, TypeVar, Callable, Optional, cast

import grpclib.exceptions
from grpclib import Status
from grpclib.client import Channel
from grpclib.config import Configuration
from grpclib.exceptions import StreamTerminatedError
from betterproto.grpc.grpclib_client import ServiceStub
from grpclib.events import SendMessage, SendRequest, RecvInitialMetadata, listen

from qm.utils.async_utils import run_async
from qm.utils.general_utils import is_debug
from qm.api.models.server_details import ConnectionDetails
from qm.exceptions import QMTimeoutError, QMConnectionError

logger = logging.getLogger(__name__)

Ret = TypeVar("Ret")
P = ParamSpec("P")


class GatewayNotImplementedError(NotImplementedError):
    pass


def _check_stream_terminated_is_timeout(func_start_time: float, func_timeout: float) -> bool:
    """
    When the function's timeout is reached, grpclib cancels the internally created future (distinct from the one created
    in the SDK). This cancellation prompts the server to reset the stream, triggering a "Stream reset by remote party"
    event and resulting in a StreamTerminatedError.
    This function verifies that the StreamTerminatedError is due to a timeout by ensuring that the function's
    runtime is approximately equal to the specified timeout.
    """
    func_end_time = time.time()
    func_runtime = func_end_time - func_start_time
    # Check if the function's runtime is within 5% of the specified timeout
    return math.isclose(func_runtime, func_timeout, rel_tol=0.05)


def connection_error_handle_decorator(func: Callable[P, Ret]) -> Callable[P, Ret]:
    @functools.wraps(func)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> Ret:
        func_start_time = time.time()
        try:
            return func(*args, **kwargs)
        except grpclib.exceptions.GRPCError as e:
            if is_debug():
                logger.exception("Encountered connection error from QOP")
            if e.status == Status.UNIMPLEMENTED:
                raise GatewayNotImplementedError(
                    f"Encountered connection error from QOP: details: {e.message}, status: {e.status}"
                ) from e
            raise QMConnectionError(
                f"Encountered connection error from QOP: details: {e.message}, status: {e.status}"
            ) from e
        except asyncio.TimeoutError as e:
            if is_debug():
                logger.exception(f"Timeout reached while running '{func.__name__}'")
            raise QMTimeoutError(f"Timeout reached while running '{func.__name__}'") from e
        except StreamTerminatedError as e:
            if is_debug():
                logger.exception(f"Stream terminated while running '{func.__name__}'")
            if "timeout" in kwargs:
                if _check_stream_terminated_is_timeout(func_start_time, cast(float, kwargs["timeout"])):
                    raise QMTimeoutError(
                        f"Stream terminated, most likely due to a timeout while running '{func.__name__}'"
                    ) from e
            raise e

    return wrapped


T = TypeVar("T")
StubType = TypeVar("StubType", bound=ServiceStub)


def connection_error_handle() -> Callable[[Type[T]], Type[T]]:
    def decorate(cls: Type[T]) -> Type[T]:
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)):
                setattr(cls, attr, connection_error_handle_decorator(getattr(cls, attr)))
        return cls

    return decorate


async def create_channel(connection_details: ConnectionDetails) -> Channel:
    return Channel(
        host=connection_details.host,
        port=connection_details.port,
        ssl=connection_details.ssl_context,
        config=Configuration(
            http2_connection_window_size=connection_details.max_message_size,
            http2_stream_window_size=connection_details.max_message_size,
        ),
    )


class BaseApi(Generic[StubType], metaclass=ABCMeta):
    def __init__(self, connection_details: ConnectionDetails):
        self._connection_details = connection_details

        self._channel = run_async(create_channel(self._connection_details))
        self._stub: StubType = self._stub_class(self._channel)

        if self._connection_details.debug_data:
            self._create_debug_data_event()

        self._create_add_headers_event()
        self._timeout: Optional[float] = self._connection_details.timeout

    @property
    def connection_details(self) -> ConnectionDetails:
        return self._connection_details

    @property
    @abstractmethod
    def _stub_class(self) -> Type[StubType]:
        pass

    def _create_debug_data_event(self) -> None:
        async def intercept_response(event: RecvInitialMetadata) -> None:
            assert self._connection_details.debug_data is not None

            metadata = event.metadata
            logger.debug(f"Collected response metadata: {json.dumps(dict(metadata), indent=4)}")
            self._connection_details.debug_data.append(metadata)

        async def send_request_debug(event: SendRequest) -> None:
            logger.debug("-----------request start-----------")
            logger.debug("   ---    request headers    ---   ")
            logger.debug(f"method:       {event.method_name}")
            logger.debug(f"metadata:     {json.dumps(dict(event.metadata), indent=4)}")
            logger.debug(f"content type: {event.content_type}")
            if event.deadline:
                deadline = event.deadline.time_remaining()
            else:
                deadline = None
            logger.debug(f"deadline:     {deadline}")

        async def send_message_debug(event: SendMessage) -> None:
            logger.debug("   ---    request message    ---   ")
            try:
                logger.debug(f"message:      {event.message.to_json(4)}")
            except TypeError:
                pass
            logger.debug("------------end request------------")

        listen(self._channel, RecvInitialMetadata, intercept_response)
        listen(self._channel, SendRequest, send_request_debug)
        listen(self._channel, SendMessage, send_message_debug)

    def _create_add_headers_event(self) -> None:
        async def add_headers(event: SendRequest) -> None:
            event.metadata.update(self._connection_details.headers)

        listen(self._channel, SendRequest, add_headers)

    @property
    def channel(self) -> Channel:
        return self._channel

    def __del__(self) -> None:
        self.channel.close()
